import base64
import json
import pickle
from django.apps.registry import Apps
from django.db.migrations import RunPython
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from garnett.utils import get_current_language_code
from typing import Callable, Dict, List


"""
These methods convert a string to and from a safe base64 encoded version of text.
This helps with database encoding
"""


def safe_encode(value: str) -> str:
    return base64.b64encode(pickle.dumps(value)).decode("ascii")


def safe_decode(value: str) -> str:
    return pickle.loads(base64.urlsafe_b64decode(value))


def _get_migrate_function(
    app_label: str,
    model_fields: Dict[str, List[str]],
    update: Callable[[str, str], str],
    ts_all: bool = False,
) -> Callable[[Apps, BaseDatabaseSchemaEditor], None]:
    """Generate a migration function given an update function for each value

    update is a function taking current language and old value and returning a new value
    """

    def migrate(apps: Apps, schema_editor: BaseDatabaseSchemaEditor) -> None:
        current_lang = get_current_language_code()

        for model_name, fields in model_fields.items():
            updated = []
            model = apps.get_model(app_label, model_name)
            for item in model.objects.all():
                for field_name in fields:
                    # Set new value retrieved from update function
                    if ts_all:
                        field_name = f"{field_name}_tsall"
                    value = getattr(item, field_name)
                    setattr(item, field_name, update(current_lang, value))
                updated.append(item)

            # Bulk update only the required fields
            model.objects.bulk_update(updated, fields)

    return migrate


# Part 1


def update_safe_encode_content_forwards(current_lang: str, value: str) -> str:
    value = safe_encode(value)
    return json.dumps({current_lang: value})


def update_safe_encode_content_backwards(current_lang: str, value: str) -> str:
    value = json.loads(value)
    return safe_decode(value.get(current_lang, safe_encode("")))


def step_1_safe_encode_content(
    app_label: str, model_fields: Dict[str, List[str]]
) -> RunPython:
    """Generate a migration operation to safely store text content prior to field migration
    This is needed as standard json is not escaped according to Postgres (and maybe other database) requirements.
    For example, Postgres requires single quotes (') to be double escaped: ('')
    For more info see here: https://stackoverflow.com/questions/35677204/psql-insert-json-with-double-quotes-inside-strings

    Args:
        app_label: label for django app the migration is in
        model_fields: mapping of model name to list of text fields to migrate

    Returns:
        RunPython migration operation
    """

    return RunPython(
        code=_get_migrate_function(
            app_label, model_fields, update_safe_encode_content_forwards
        ),
        reverse_code=_get_migrate_function(
            app_label, model_fields, update_safe_encode_content_backwards, ts_all=True
        ),
    )


# Part 2


def update_safe_prepare_translations_forwards(current_lang: str, value: dict) -> dict:
    value = safe_decode(value.get(current_lang, safe_encode("")))
    return {current_lang: value}


def update_safe_prepare_translations_backwards(current_lang: str, value: dict) -> dict:
    value = safe_encode(value.get(current_lang, ""))
    return {current_lang: value}


def step_2_safe_prepare_translations(
    app_label: str, model_fields: Dict[str, List[str]]
) -> RunPython:
    """Generate a migration operation to prepare text fields for being translated

    Args:
        app_label: label for django app the migration is in
        model_fields: mapping of model name to list of text fields to migrate

    Returns:
        RunPython migration operation
    """

    return RunPython(
        code=_get_migrate_function(
            app_label,
            model_fields,
            update_safe_prepare_translations_forwards,
            ts_all=True,
        ),
        reverse_code=_get_migrate_function(
            app_label,
            model_fields,
            update_safe_prepare_translations_backwards,
            ts_all=True,
        ),
    )



from typing import List, Union, Optional

import langcodes.tag_parser
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from langcodes import Language

from garnett.context import _ctx_language, _ctx_force_blank


def lang_param():
    return getattr(settings, "GARNETT_QUERY_PARAMETER_NAME", "glang")


def get_default_language():
    setting = getattr(settings, "GARNETT_DEFAULT_TRANSLATABLE_LANGUAGE", "en-AU")
    if callable(setting):
        default = setting()
    else:
        default = setting

    if isinstance(default, Language):
        return default
    elif isinstance(default, str):
        return Language.get(default)
    else:
        raise ImproperlyConfigured(
            "GARNETT_DEFAULT_TRANSLATABLE_LANGUAGE must be a string or callable that returns a string or `Language` object"
        )


def codes_to_langs(content: dict) -> dict:
    return {Language.get(lang): text for lang, text in content.items()}


def get_property_name() -> str:
    return getattr(settings, "GARNETT_TRANSLATIONS_PROPERTY_NAME", "translations")


def normalise_language_code(langcode):
    return Language.get(langcode).to_tag()


def normalise_language_codes(value):
    return {normalise_language_code(lang): val for lang, val in value.items()}


def is_valid_language(language: Union[str, Language]) -> bool:
    if isinstance(language, Language):
        language = language
    if isinstance(language, str):
        try:
            language = Language.get(language)
        except langcodes.tag_parser.LanguageTagError:
            return False
    return language in get_languages()


def get_current_language() -> Language:
    lang = _ctx_language.get(None)
    if not lang:
        return get_default_language()
    return lang


def get_current_language_code() -> str:
    return get_current_language().to_tag()


def get_current_blank_override() -> bool:
    return _ctx_force_blank.get(False)


def get_safe_language(lang_code: str) -> Optional[Language]:
    """Return language if language for lang code exists, otherwise none"""
    try:
        return Language.get(lang_code)
    except langcodes.tag_parser.LanguageTagError:
        return None


def validate_language_list(langs) -> List[Language]:
    """
    Validate and clean a potential list of languages.
    This may return an empty list if the provided languages are invalid
    """
    if type(langs) is not list:
        return []

    languages = []
    for lang_code in langs:
        if language := get_safe_language(lang_code):
            languages.append(language)

    if languages:
        return languages


def get_languages() -> List[Language]:
    langs = getattr(
        settings, "GARNETT_TRANSLATABLE_LANGUAGES", [get_default_language()]
    )
    if callable(langs):
        langs = langs()

    languages = validate_language_list(langs)

    if not languages:
        raise ImproperlyConfigured(
            "GARNETT_TRANSLATABLE_LANGUAGES must be a list of languages or a callable that returns a list of languages"
        )
    return languages
